from __future__ import annotations

from typing import Optional
from unittest import TestCase

from stereotype import Model, Missing, ValidationError, ConversionError, BoolField, IntField, ConfigurationError, \
    FloatField, StrField


class BoolModel(Model):
    req: bool
    opt: Optional[bool]
    req_def: bool = True
    opt_def: Optional[bool] = None


class TestBooleanField(TestCase):
    def test_empty(self):
        model = BoolModel()
        self.assertIs(Missing, model.req)
        self.assertIs(Missing, model.opt)
        self.assertIs(True, model.req_def)
        self.assertIs(None, model.opt_def)
        self.assertFalse(model.req)
        self.assertEqual('<BoolModel {req=Missing, opt=Missing, req_def=True, opt_def=None}>', str(model))
        self.assertEqual('<BoolModel {req=Missing, opt=Missing, req_def=True, opt_def=None}>', repr(model))
        self.assertEqual({'req_def': True, 'opt_def': None}, model.serialize())
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual('req: This field is required', str(ctx.exception))
        self.assertEqual({'req': ['This field is required'], 'opt': ['This field is required']}, ctx.exception.errors)

        self.assertEqual('<Field req of type bool, required>', repr(BoolModel.__fields__[0]))

    def test_special(self):
        class BoolSpecialModel(Model):
            plain: bool = False
            hidden: Optional[bool] = BoolField(hide_none=True)
            changed: bool = BoolField(primitive_name='other', hide_false=True)
            weird: Optional[bool] = BoolField(primitive_name='input', to_primitive_name='output', default=True)

        model = BoolSpecialModel({'hidden': None, 'changed': 'ignore', 'other': True, 'output': False, 'weird': False})
        self.assertIs(False, model.plain)
        self.assertIs(None, model.hidden)
        self.assertIs(True, model.changed)
        self.assertIs(True, model.weird)
        self.assertEqual({'plain': False, 'other': True, 'output': True}, model.serialize())
        model.validate()

        model.hidden = True
        model.changed = False
        model.weird = False
        self.assertEqual({'plain': False, 'hidden': True, 'output': False}, model.serialize())
        model.validate()

        model.changed = None
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual({'other': ['This field is required']}, ctx.exception.errors)

        self.assertEqual('<Field plain of type bool, default=<False>>', repr(BoolSpecialModel.__fields__[0]))

    def test_bad_values(self):
        model = BoolModel({'req': None, 'opt': 'No', 'opt_def': 'true'})
        self.assertIs(None, model.req)
        self.assertIs(False, model.opt)
        self.assertIs(True, model.req_def)
        self.assertIs(True, model.opt_def)
        self.assertEqual('<BoolModel {req=None, opt=False, req_def=True, opt_def=True}>', repr(model))
        self.assertEqual({'req': None, 'opt': False, 'req_def': True, 'opt_def': True}, model.to_primitive())
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual('req: This field is required', str(ctx.exception))
        self.assertEqual({'req': ['This field is required']}, ctx.exception.errors)

    def test_bad_types(self):
        with self.assertRaises(ConversionError) as ctx:
            BoolModel({'req': 'bad', 'opt': 4, 'opt_def': 'true'})
        self.assertEqual('req: Value must be a boolean or a true/false/yes/no string value', str(ctx.exception))
        with self.assertRaises(ConversionError) as ctx:
            BoolModel({'opt': 4, 'opt_def': 'false'})
        self.assertEqual({'opt': ['Value must be a boolean or a true/false/yes/no string value']}, ctx.exception.errors)

    def test_bad_default(self):
        class NotBool(Model):
            value: bool = 42

        with self.assertRaises(ConfigurationError) as ctx:
            NotBool()
        self.assertEqual('Value `42` used as field default must be of type bool', str(ctx.exception))


class IntModel(Model):
    normal: int = IntField(default=42, primitive_name='norm')
    min: int = IntField(min_value=3)
    max: Optional[int] = IntField(max_value=int(2e10))
    min_max: int = IntField(min_value=5, max_value=10, default=lambda: 8)


class TestIntField(TestCase):
    def test_min_max_validation(self):
        model = IntModel({'norm': -5, 'min': 2, 'max': int(3e10), 'min_max': '17'})
        self.assertEqual(-5, model.normal)
        self.assertEqual(2, model.min)
        self.assertEqual(int(3e10), model.max)
        self.assertEqual(17, model.min_max)
        self.assertEqual(17, model['min_max'])
        with self.assertRaises(KeyError) as ctx:
            self.failIf(model['serialize'])
        self.assertEqual("'serialize'", str(ctx.exception))
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual({
            'min': ['Must be at least 3'],
            'max': ['Must be at most 20000000000'],
            'min_max': ['Must be between 5 and 10'],
        }, ctx.exception.errors)
        self.assertEqual({'norm': -5, 'min': 2, 'max': 30000000000, 'min_max': 17}, model.serialize())

        self.assertEqual('<Field normal of type int, default=<42>, primitive name norm>', repr(IntModel.__fields__[0]))

    def test_conversion_errors(self):
        with self.assertRaises(ConversionError) as ctx:
            IntModel({'min': 2.5, 'max': 100, 'min_max': 'abc'})
        self.assertEqual({'min': ['Numeric value 2.5 is not an integer']}, ctx.exception.errors)
        with self.assertRaises(ConversionError) as ctx:
            IntModel({'max': 100, 'min_max': 'abc'})
        self.assertEqual({'min_max': ["Value 'abc' is not an integer number"]}, ctx.exception.errors)
        with self.assertRaises(ConversionError) as ctx:
            IntModel({'max': '1.5'})
        self.assertEqual({'max': ["Value '1.5' is not an integer number"]}, ctx.exception.errors)

    def test_none_and_defaults(self):
        model = IntModel({'min': 4, 'max': None})
        self.assertEqual(42, model.normal)
        self.assertEqual(4, model.min)
        self.assertIsNone(model.max)
        self.assertEqual(8, model.min_max)
        model.validate()
        self.assertEqual({'norm': 42, 'min': 4, 'max': None, 'min_max': 8}, model.serialize())

    def test_bad_default(self):
        class NotInt(Model):
            field: int = 4.2

        with self.assertRaises(ConfigurationError) as ctx:
            NotInt()
        self.assertEqual('Value `4.2` used as field default must be of type int', str(ctx.exception))

        class NotOptional(Model):
            field: int = None

        with self.assertRaises(ConfigurationError) as ctx:
            NotOptional()
        self.assertEqual('Field `field` is not Optional and cannot use None as default', str(ctx.exception))

    def test_hidden_figures(self):
        class Hidden(Model):
            none: Optional[int] = IntField(hide_none=True)
            zero: Optional[int] = IntField(hide_zero=True)

        self.assertEqual({'none': 0}, Hidden({'none': 0, 'zero': 0}).serialize())
        self.assertEqual({'zero': None}, Hidden({'none': None, 'zero': None}).serialize())


class FloatModel(Model):
    normal: Optional[float] = None
    min: float = FloatField(min_value=4, to_primitive_name='bottom')
    max: Optional[float] = FloatField(max_value=47.42, default=42., primitive_name='top', to_primitive_name='max')
    min_max: float = FloatField(min_value=1.5, max_value=4, default=lambda: 3.5, hide_none=True)

    def validate_min(self, value: int, _):
        # Would fail if executed with None
        if value % 2 == 0:
            raise ValueError('Must be an odd number')

    def validate_max(self, value: int, _):
        # Executed even with None because max is optional
        if value is None and self.normal is None:
            raise ValueError('Must be set if normal is omitted')


class TestFloatField(TestCase):
    def test_min_max_validation(self):
        FloatModel({'normal': 1, 'min': 47, 'min_max': 1.5}).validate()
        model = FloatModel({'normal': -5.6, 'min': 2, 'top': '50.42', 'min_max': '8', 'max': 'ignore me'})
        self.assertEqual(-5.6, model.normal)
        self.assertEqual(2, model.min)
        self.assertEqual(50.42, model.max)
        self.assertEqual(8, model.min_max)
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual({
            'min': ['Must be at least 4', 'Must be an odd number'],
            'top': ['Must be at most 47.42'],
            'min_max': ['Must be between 1.5 and 4'],
        }, ctx.exception.errors)
        self.assertEqual({'normal': -5.6, 'bottom': 2., 'max': 50.42, 'min_max': 8.}, model.serialize())

        self.assertEqual('<Field normal of type Optional[float], default=<None>>', repr(FloatModel.__fields__[0]))

    def test_custom_validation(self):
        model = FloatModel({'min': 4, 'top': None})
        self.assertEqual(None, model.normal)
        self.assertEqual(4., model.min)
        self.assertEqual(None, model.max)
        self.assertEqual(3.5, model.min_max)
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual({
            'min': ['Must be an odd number'],
            'top': ['Must be set if normal is omitted'],
        }, ctx.exception.errors)
        self.assertEqual({'normal': None, 'bottom': 4., 'max': None, 'min_max': 3.5}, model.serialize())
        model.normal = model.min = 5.
        model.validate()

    def test_none_and_defaults(self):
        model = FloatModel({'min': 5})
        self.assertIsNone(None, model.normal)
        self.assertEqual(5., model.min)
        self.assertEqual(42., model.max)
        self.assertEqual(3.5, model.min_max)
        self.assertTrue(isinstance(model.min, float))
        model.validate()
        self.assertEqual({'normal': None, 'bottom': 5., 'max': 42., 'min_max': 3.5}, model.serialize())

        model.normal = 4.5
        model.min = model.max = model.min_max = None
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual({'min': ['This field is required'], 'min_max': ['This field is required']},
                         ctx.exception.errors)

    def test_mismatched_field(self):
        class Mismatch(Model):
            my_name: int = FloatField(min_value=1)

        with self.assertRaises(ConfigurationError) as ctx:
            Mismatch({'my_name': 5})
        self.assertEqual('Annotations for <Field my_name of type int, required> require custom field type IntField, '
                         'got FloatField', str(ctx.exception))

    def test_hidden_figures(self):
        class Hidden(Model):
            none: Optional[float] = FloatField(hide_none=True, default=None)
            zero: Optional[float] = FloatField(hide_zero=True, default=0.0)

        self.assertEqual({}, Hidden({}).serialize())
        self.assertEqual({}, Hidden({'none': None, 'zero': 0}).serialize())
        self.assertEqual({'none': 0, 'zero': None}, Hidden({'none': 0, 'zero': None}).serialize())


class TestStrField(TestCase):
    def test_validation(self):
        class StrModel(Model):
            normal: str
            non_empty: Optional[str] = StrField(min_length=1)
            min: str = StrField(min_length=3)
            max: str = StrField(max_length=1)
            min_max: Optional[str] = StrField(min_length=1, max_length=6)
            exact: str = StrField(min_length=3, max_length=3)
            choices: Optional[str] = StrField(choices=('a', 'bb', 'ccc'))

        StrModel({
            'normal': 'a',
            'non_empty': 'abc',
            'min': 'abc',
            'max': '',
            'min_max': 'a',
            'exact': 'abc',
            'choices': 'bb',
        }).validate()

        model = StrModel({
            'normal': 4,
            'non_empty': '',
            'min': 'x',
            'max': '123',
            'min_max': '',
            'exact': True,
            'choices': 'c',
        })
        self.assertEqual('4', model.normal)
        self.assertEqual('', model.non_empty)
        self.assertEqual('x', model.min)
        self.assertEqual('123', model.max)
        self.assertEqual('', model.min_max)
        self.assertEqual('True', model.exact)
        self.assertEqual('c', model.choices)
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual({
            'non_empty': ['This value cannot be empty'],
            'min': ['Must be at least 3 characters long'],
            'max': ['Must be at most 1 character long'],
            'min_max': ['Must be 1 to 6 characters long'],
            'exact': ['Must be exactly 3 characters long'],
            'choices': ['Must be one of: a, bb, ccc'],
        }, ctx.exception.errors)

        self.assertEqual('<Field non_empty of type Optional[str], required>', repr(StrModel.__fields__[1]))

    def test_bad_validation_combo(self):
        with self.assertRaises(ConfigurationError) as ctx:
            class LengthChoices(Model):
                unused: LengthChoices
                bad: str = StrField(min_length=1, choices=('x', 'y', 'zz'))
        self.assertEqual('Cannot use min_length or max_length together with choices', str(ctx.exception))

    def test_hide_empty(self):
        class Hidden(Model):
            none: Optional[str] = StrField(hide_none=True)
            empty: Optional[str] = StrField(hide_empty=True)

        self.assertEqual({'empty': None}, Hidden({'none': None, 'empty': None}).serialize())
        self.assertEqual({'none': ''}, Hidden({'none': '', 'empty': ''}).serialize())



from __future__ import annotations

from typing import Any, Optional, Type, Iterable, Tuple, Dict, cast

from stereotype.fields.base import Field
from stereotype.model import Model
from stereotype.utils import Missing


class ModelField(Field):
    __slots__ = Field.__slots__ + ('type',)
    atomic = False

    def __init__(self, *, default: Any = Missing, hide_none: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing):
        super().__init__(default=default, hide_none=hide_none,
                         primitive_name=primitive_name, to_primitive_name=to_primitive_name)

        self.type: Type[Model] = cast(Type[Model], NotImplemented)
        self.native_validate = self.validate

    def validate(self, value: Any, context: dict) -> Iterable[Tuple[Tuple[str, ...], str]]:
        yield from value.validation_errors(context)

    def type_config_from(self, field: ModelField):
        super().type_config_from(field)
        self.type = field.type

    def convert(self, value: Any) -> Any:
        if value is Missing:
            if self.required:
                return Missing
            return self.default if self.default_factory is None else self.default_factory()
        if value is None:
            return None
        if isinstance(value, Model):
            if not isinstance(value, self.type):
                raise TypeError(f'Supplied type {type(value).__name__}, needs a mapping or {self.type.__name__}')
            return value
        return self.type(value)

    def copy_value(self, value: Any) -> Any:
        if value is None or value is Missing:
            return value
        return value.copy(deep=True)

    def to_primitive(self, value: Any) -> Any:
        if value is None or value is Missing:
            return value
        return value.to_primitive()

    @property
    def type_repr(self):
        return self.type.__name__ if self.type is not NotImplemented else '(unknown)'


class DynamicModelField(Field):
    __slots__ = Field.__slots__ + ('types', 'type_map')
    atomic = False
    type = Model

    def __init__(self, *, default: Any = Missing, hide_none: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing):
        super().__init__(default=default, hide_none=hide_none,
                         primitive_name=primitive_name, to_primitive_name=to_primitive_name)

        self.types: Tuple[Type[Model], ...] = NotImplemented
        self.type_map: Dict[str, Type[Model]] = NotImplemented
        self.native_validate = self.validate

    def validate(self, value: Any, context: dict) -> Iterable[Tuple[Tuple[str, ...], str]]:
        yield from value.validation_errors(context)

    def type_config_from(self, field: DynamicModelField):
        super().type_config_from(field)
        self.types = field.types
        self.type_map = field.type_map

    def convert(self, value: Any) -> Any:
        if value is Missing:
            if self.required:
                return Missing
            return self.default if self.default_factory is None else self.default_factory()
        is_model = isinstance(value, Model)
        if is_model and not isinstance(value, self.types):
            raise TypeError(f'Expected {self.type_repr}, got {type(value).__name__}')
        if is_model or value is None:
            return value

        try:
            value_type = value['type']
        except TypeError:
            raise TypeError(f'Expected a mapping with a `type` field, got type {type(value).__name__}')
        except KeyError:
            raise TypeError('Expected a mapping with a `type` field, got no `type` field')
        type_cls = self.type_map.get(value_type, None)
        if type_cls is None:
            raise TypeError(f'Got a mapping with unsupported `type` {value_type!r}')
        return type_cls(value)

    def copy_value(self, value: Any) -> Any:
        if value is None or value is Missing:
            return value
        return value.copy(deep=True)

    def to_primitive(self, value: Any) -> Any:
        if value is None or value is Missing:
            return value
        result = value.to_primitive()
        result['type'] = value.type
        return result

    @property
    def type_repr(self):
        return f'Union[{", ".join(option.__name__ for option in self.types)}]'



from __future__ import annotations

from threading import Lock
from typing import List, Set, Optional, Any, Tuple

from stereotype.utils import ConfigurationError


class Role:
    __slots__ = ('code', 'name', 'empty_by_default')

    def __init__(self, name: str, empty_by_default: bool = False):
        self.name = name
        self.empty_by_default = empty_by_default
        with _roles_lock:
            self.code = len(_roles)
            _roles.append(self)

    def __repr__(self):
        return f'<Role {self.name}, empty_by_default={self.empty_by_default}, code={self.code}>'

    def __hash__(self):
        return self.code

    def __eq__(self, other):
        return type(self) == type(other) and self.code == other.code

    def whitelist(self, *fields, override_parents: bool = False):
        return RequestedRoleFields(self, fields, is_whitelist=True, override_parents=override_parents)

    def blacklist(self, *fields, override_parents: bool = False):
        return RequestedRoleFields(self, fields, is_whitelist=False, override_parents=override_parents)


_roles: List[Role] = []
_roles_lock = Lock()

DEFAULT_ROLE = Role('default')


class FinalizedRoleFields:
    __slots__ = ('role', 'fields')

    def __init__(self, role: Role, fields: Optional[Set[str]] = None):
        self.role = role
        self.fields = fields or set()

    def update_requested(self, other: RequestedRoleFields, all_field_names: Set[str], field_names: Set[str]):
        assert self.role == other.role
        if other.override_parents:
            initial = set() if other.is_whitelist else all_field_names
        else:
            initial = self.fields
        if other.is_whitelist:
            self.fields = initial | other.fields
        else:
            self.fields = (initial | field_names) - other.fields


class RequestedRoleFields:
    __slots__ = ('role', 'fields', 'is_whitelist', 'override_parents')

    def __init__(self, role: Role, fields, is_whitelist: bool, override_parents: bool):
        self.fields, non_descriptors = self._collect_input_fields(fields)
        if non_descriptors:
            raise ConfigurationError(f'Role blacklist/whitelist needs member descriptors (e.g. cls.my_field), '
                                     f'got {non_descriptors[0]!r}')
        self.role = role
        self.is_whitelist = is_whitelist
        self.override_parents = override_parents

    def _collect_input_fields(self, fields) -> Tuple[Set[str], List[Any]]:
        field_names: Set[str] = set()
        non_descriptors: List[Any] = []
        for field in fields:
            if type(field).__name__ == 'member_descriptor':
                field_names.add(field.__name__)
            elif isinstance(field, property):
                field_names.add(field.fget.__name__)
            else:
                non_descriptors.append(field)
        return field_names, non_descriptors



from __future__ import annotations

from typing import Optional, Union, Type, Set
from unittest import TestCase

from stereotype import Model, Missing, ValidationError, ConversionError, ModelField, DynamicModelField, \
    ConfigurationError
from tests.common import Leaf


class Fake(object):
    type = 'fake'


class Branch(Model):
    type = 'branch'

    leaf: Leaf
    branch: Optional[Branch] = ModelField(default=None, hide_none=True, primitive_name='sub-branch')


class Trunk(Model):
    left: Optional[Union[Branch, Leaf]] = None
    right: Union[Leaf, None, Branch] = DynamicModelField(default=lambda: None, primitive_name='right')

    def validate_right(self, value: Optional[Union[Branch, Leaf]], _):
        if isinstance(value, Branch) and value.leaf.color == 'green':
            raise ValueError('Trees cannot have green leaves on the right side')


class Root(Model):
    depth: int
    trunk: Optional[Trunk] = Trunk


class TestModelType(TestCase):
    def test_empty(self):
        model = Branch()
        self.assertIs(Missing, model.leaf)
        self.assertIs(None, model.branch)
        self.assertEqual('<Branch {leaf=Missing, branch=None}>', repr(model))
        self.assertEqual({}, model.serialize())
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual('leaf: This field is required', str(ctx.exception))
        self.assertEqual({'leaf': ['This field is required']}, ctx.exception.errors)

        self.assertEqual('<Field leaf of type Leaf, required>', repr(Branch.__fields__[0]))

    def test_simple(self):
        model = Branch({'leaf': {'color': 'red'}})
        self.assertEqual(Leaf({'color': 'red'}), model.leaf)
        self.assertIs(None, model.branch)
        self.assertEqual('<Branch {leaf=Leaf, branch=None}>', repr(model))
        self.assertEqual({'leaf': {'color': 'red'}}, model.serialize())
        model.validate()

    def test_recursive(self):
        model = Branch({'leaf': Leaf({'color': 'yellow'}), 'sub-branch': {'leaf': {}}})
        self.assertEqual(Leaf({'color': 'yellow'}), model.leaf)
        self.assertEqual(Branch({'leaf': {'color': 'green'}}), model.branch)
        self.assertEqual('green', model.branch.leaf.color)
        self.assertEqual('<Branch {leaf=Leaf, branch=Branch}>', repr(model))
        self.assertEqual({'leaf': {'color': 'yellow'}, 'sub-branch': {'leaf': {'color': 'green'}}}, model.serialize())
        model.validate()

    def test_bad_type(self):
        with self.assertRaises(ConversionError) as ctx:
            Branch({'leaf': 'yellow'})
        self.assertEqual({'leaf': ['Supplied type str, needs a mapping']}, ctx.exception.errors)
        with self.assertRaises(ConversionError) as ctx:
            Branch({'leaf': Branch()})
        self.assertEqual({'leaf': ['Supplied type Branch, needs a mapping or Leaf']}, ctx.exception.errors)
        with self.assertRaises(ValidationError) as ctx:
            Branch({'leaf': None}).validate()
        self.assertEqual({'leaf': ['This field is required']}, ctx.exception.errors)

    def test_default_model(self):
        model = Root()
        self.assertIs(Missing, model.depth)
        self.assertEqual(Trunk(), model.trunk)
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual('depth: This field is required', str(ctx.exception))
        model.depth = 2
        model.validate()
        self.assertIsNone(model.trunk.left)
        model.trunk = None
        self.assertEqual({'depth': 2, 'trunk': None}, model.serialize())


class TestDynamicModelField(TestCase):
    def test_empty(self):
        model = Trunk()
        self.assertIsNone(model.left)
        self.assertIsNone(model.right)
        model.validate()
        self.assertEqual({'left': None, 'right': None}, model.to_primitive())
        self.assertEqual('<Field left of type Optional[Union[Branch, Leaf]], default=<None>>',
                         repr(Trunk.__fields__[0]))

    def test_mixed(self):
        model = Trunk({'left': {'type': 'branch', 'leaf': {'color': 'yellow'}}, 'right': Leaf({'color': 'red'})})
        self.assertEqual(Branch({'leaf': {'color': 'yellow'}}), model.left)
        self.assertEqual(Leaf({'color': 'red'}), model.right)
        model.validate()
        self.assertEqual({'left': {'type': 'branch', 'leaf': {'color': 'yellow'}},
                          'right': {'type': 'leaf', 'color': 'red'}}, model.to_primitive())
        model.right = Branch({'leaf': {}})
        self.assertEqual({'left': {'type': 'branch', 'leaf': {'color': 'yellow'}},
                          'right': {'type': 'branch', 'leaf': {'color': 'green'}}}, model.to_primitive())
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual({'right': ['Trees cannot have green leaves on the right side']}, ctx.exception.errors)

    def test_bad_configuration_non_model(self):
        class Bad(Model):
            field: Union[Leaf, Fake]

        with self.assertRaises(ConfigurationError) as ctx:
            Bad()
        self.assertEqual('Union Model fields can only be Optional or Union of Model subclass types, '
                         'got typing.Union[tests.common.Leaf, tests.test_model_fields.Fake]', str(ctx.exception))

    def test_bad_configuration_no_type(self):
        class Worse(Model):
            field: Union[Leaf, Trunk, None]

        with self.assertRaises(ConfigurationError) as ctx:
            Worse()
        self.assertEqual('Model Trunk used in a dynamic model field '
                         'typing.Union[tests.common.Leaf, tests.test_model_fields.Trunk] '
                         'but does not define a non-type-annotated string `type` field', str(ctx.exception))

    def test_bad_configuration_type_is_a_field(self):
        class Weird(Model):
            type: str = 'rogue annotation'

        class Horrible(Model):
            field: Union[Weird, Leaf]

            @classmethod
            def resolve_extra_types(cls) -> Set[Type[Model]]:
                return {Weird}

        with self.assertRaises(ConfigurationError) as ctx:
            Horrible()
        self.assertEqual(
            "Model Weird used in a dynamic model field "
            "typing.Union[tests.test_model_fields.TestDynamicModelField.test_bad_configuration_type_is_a_field."
            "<locals>.Weird, tests.common.Leaf] "
            "but it's `type` field has a type annotation making it a field, must be an attribute",
            str(ctx.exception),
        )

    def test_bad_configuration_type_not_a_string(self):
        class Nonsense(Model):
            type = 5

        class Atrocity(Model):
            field: Union[Branch, Nonsense]

            @classmethod
            def resolve_extra_types(cls) -> Set[Type[Model]]:
                return {Branch, Nonsense, Leaf}

        with self.assertRaises(ConfigurationError) as ctx:
            Atrocity()
        self.assertEqual("Model Nonsense used in a dynamic model field "
                         "typing.Union[tests.test_model_fields.Branch, tests.test_model_fields.TestDynamicModelField."
                         "test_bad_configuration_type_not_a_string.<locals>.Nonsense] "
                         "but it's `type` field 5 is not a string", str(ctx.exception))

    def test_bad_configuration_type_conflict(self):
        class AnotherLeaf(Leaf):
            pass

        class CrimeAgainstHumanity(Model):
            field: Union[Leaf, Branch, AnotherLeaf]

            @classmethod
            def resolve_extra_types(cls) -> Set[Type[Model]]:
                return {AnotherLeaf}

        with self.assertRaises(ConfigurationError) as ctx:
            CrimeAgainstHumanity()
        self.assertEqual("Conflicting dynamic model field types in typing.Union["
                         "tests.common.Leaf, tests.test_model_fields.Branch, tests.test_model_fields."
                         "TestDynamicModelField.test_bad_configuration_type_conflict.<locals>.AnotherLeaf"
                         "]: Leaf vs AnotherLeaf", str(ctx.exception))

    def test_bad_value(self):
        class Dynamic(Model):
            field: Union[Leaf, Branch]

        with self.assertRaises(ConversionError) as ctx:
            Dynamic({'field': Root({'depth': 1})})
        self.assertEqual('field: Expected Union[Leaf, Branch], got Root', str(ctx.exception))

        with self.assertRaises(ConversionError) as ctx:
            Dynamic({'field': []})
        self.assertEqual('field: Expected a mapping with a `type` field, got type list', str(ctx.exception))

        with self.assertRaises(ConversionError) as ctx:
            Dynamic({'field': {'color': 'no type'}})
        self.assertEqual('field: Expected a mapping with a `type` field, got no `type` field', str(ctx.exception))

        with self.assertRaises(ConversionError) as ctx:
            Dynamic({'field': {'type': 'nest', 'color': 'no type'}})
        self.assertEqual("field: Got a mapping with unsupported `type` 'nest'", str(ctx.exception))

        with self.assertRaises(ValidationError) as ctx:
            Dynamic().validate()
        self.assertEqual('field: This field is required', str(ctx.exception))

        with self.assertRaises(ValidationError) as ctx:
            Dynamic({'field': None}).validate()
        self.assertEqual('field: This field is required', str(ctx.exception))



from .fields.atomic import BoolField, IntField, FloatField, StrField  # noqa
from .fields.compound import ListField, DictField  # noqa
from .fields.model import ModelField, DynamicModelField  # noqa
from .fields.serializable import serializable  # noqa
from .model import Model  # noqa
from .roles import Role, DEFAULT_ROLE, RequestedRoleFields  # noqa
from .utils import Missing, StereotypeError, ConfigurationError, DataError, ValidationError, ConversionError  # noqa



from typing import Union, List, Dict, Tuple, cast


class _MissingType:
    def __bool__(self):
        return False

    def __repr__(self):
        return 'Missing'


Missing = _MissingType()


class StereotypeError(Exception):
    pass


class ConfigurationError(StereotypeError, AssertionError):
    pass


class DataError(StereotypeError):
    error_list: List[Tuple[Tuple[str, ...], str]]

    def __init__(self, errors: List[Tuple[Tuple[str, ...], str]]):
        self.error_list = errors
        super().__init__(self._error_string())

    def _error_string(self) -> str:
        for path, error in self.error_list:
            if not path:
                return error
            return f'{": ".join(path)}: {error}'
        assert self.error_list, 'Cannot create the exception without any errors'

    @property
    def errors(self) -> Dict[str, Union[List[str], dict]]:
        errors = {}
        for path, error in self.error_list:
            container = errors
            if not path:
                path = ('_global',)
            for item in path[:-1]:
                value = container.setdefault(item, {})
                if isinstance(value, list):
                    value = container[item] = cast(dict, {'_global': value})
                container = value
            array = container.setdefault(path[-1], [])
            array.append(error)
        return errors


class ValidationError(DataError, ValueError):
    pass


class ConversionError(DataError, TypeError):
    pass



from typing import List, Dict

from stereotype import Model, ListField, StrField, FloatField, ValidationError, DictField


# Keep this file synchronized with the `README.md` documentation!

class Book(Model):
    name: str
    authors: List[str] = ListField(min_length=1)
    reviews: Dict[str, float] = DictField(
        key_field=StrField(min_length=1),
        value_field=FloatField(min_value=0, max_value=5),
        default=dict,
    )


model = Book({'name': 'The Go Programming Language'})
model.authors = ['Alan A.A. Donovan', 'Brian W. Kernighan']
model.validate()
assert model.to_primitive() == {
    'name': 'The Go Programming Language',
    'authors': ['Alan A.A. Donovan', 'Brian W. Kernighan'],
    'reviews': {},
}

model = Book({
    'name': "The Hitchhiker's Guide to the Galaxy",
    'authors': ['Douglas Adams', None],
    'reviews': {'dolphins': 4.2, 'whales': -2},
})
assert model.reviews['dolphins'] == 4.2
try:
    model.validate()
    assert False, 'Validation fails due to the error below'  # pragma: no cover
except ValidationError as e:
    assert e.errors == {
        'authors': {'1': ['This field is required']},
        'reviews': {'whales': ['Must be between 0 and 5']},
    }



from __future__ import annotations

from copy import deepcopy
from typing import Any, Optional, Callable, Iterable, Tuple, TYPE_CHECKING, Dict

from stereotype.utils import Missing, ConfigurationError

if TYPE_CHECKING:  # pragma: no cover
    from stereotype.model import InputFieldConfig, OutputFieldConfig, ValidatedFieldConfig


class Field:
    __slots__ = ('name', 'required', 'allow_none', 'default', 'default_factory', 'native_validate', 'validator_method',
                 'hide_none', 'hide_empty', 'primitive_name', 'to_primitive_name', 'serializable')
    type = NotImplemented
    type_repr: str = NotImplemented
    atomic: bool = True
    empty_value = NotImplemented

    def __init__(self, *, default: Any = Missing, hide_none: bool = False, hide_empty: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing):
        """
        :param default: default value (including None) if not present in primitive data, required if omitted
        :param primitive_name: alternative field name for primitive data
        """
        from stereotype.model import Model

        # All NotImplemented *must* be updated later based on annotations
        self.name: str = NotImplemented
        self.allow_none: bool = False
        self.primitive_name: Optional[str] = primitive_name  # Missing falls back to field name
        self.to_primitive_name: Optional[str] = to_primitive_name  # Missing falls back to field name
        if primitive_name is not Missing and to_primitive_name is Missing:
            self.to_primitive_name = primitive_name

        self.native_validate: Optional[Callable[[Any, Optional[Dict]], Iterable[Tuple[Tuple[str, ...], str]]]] = None
        self.validator_method: Optional[Callable[[Model, Any, Optional[dict]], None]] = None
        self.serializable: Optional[Callable[[Model], Any]] = None

        # Only user-specifiable options are allowed as arguments to avoid user confusion
        self.required: bool = True
        self.default: Optional[Any] = None
        self.default_factory: Optional[Callable[[], Any]] = None
        if default is not Missing:
            self.set_default(default)
        self.hide_none = hide_none
        assert not (hide_empty and self.empty_value is NotImplemented), f'{type(self)} does not support hide_empty'
        self.hide_empty = hide_empty

    def set_default(self, default: Any):
        self.required = False
        if callable(default):
            self.default_factory = default
        else:
            self.default = default

    def validate(self, value: Any, context: dict) -> Iterable[Tuple[Tuple[str, ...], str]]:
        yield from ()

    def type_config_from(self, field: Field):
        self.allow_none = field.allow_none

    def fill_in_name(self, name: str):
        self.name = name
        if self.primitive_name is Missing:
            self.primitive_name = name
        if self.to_primitive_name is Missing:
            self.to_primitive_name = name

    def check_default(self):
        if self.required or self.default_factory is not None:
            pass
        elif self.default is None:
            if not self.allow_none:
                raise ConfigurationError(f'Field `{self.name}` is not Optional and cannot use None as default')
        elif not isinstance(self.default, self.type):
            raise ConfigurationError(f'Value `{self.default}` used as field default must be of type {self.type_repr}')

    def _fill_missing(self):
        if self.required:
            return Missing
        if self.default_factory is None:
            return self.default
        return self.default_factory()

    def convert(self, value: Any) -> Any:
        if value is Missing:
            return self._fill_missing()
        if value is None:
            return None
        return self.type(value)

    def to_primitive(self, value: Any) -> Any:
        return value

    def copy_field(self):
        copied = type(self)()
        for slot in type(self).__slots__:
            value = getattr(self, slot)
            if slot == 'native_validate' and value is not None:
                value = getattr(copied, value.__func__.__name__)
            setattr(copied, slot, value)
        return copied

    def copy_value(self, value: Any) -> Any:
        return value

    def make_input_config(self) -> InputFieldConfig:
        return self.name, self.primitive_name, self.convert, (None if self.atomic else self.copy_value)

    def make_validated_config(self) -> ValidatedFieldConfig:
        return (self.name, self.primitive_name or self.to_primitive_name or self.name, self.allow_none,
                self.native_validate, self.validator_method)

    def make_output_config(self) -> OutputFieldConfig:
        return (self.name, self.serializable, self.atomic, self.to_primitive,
                self.to_primitive_name, self.hide_none, self.hide_empty, self.empty_value)

    def has_validation(self) -> bool:
        return self.required or not self.allow_none or self.validator_method or self.native_validate

    def __repr__(self):
        type_repr = f'Optional[{self.type_repr}]' if self.allow_none else self.type_repr
        return (
            f'<Field{f" {self.name}" if self.name is not NotImplemented else ""} of type {type_repr}, '
            f'{"required" if self.required else f"default=<{self.default}>"}'
            f'{f", primitive name {self.primitive_name}" if self.primitive_name != self.name else ""}'
            f'{f", to output {self.to_primitive_name}" if self.to_primitive_name != self.primitive_name else ""}>'
        )


class AnyField(Field):
    __slots__ = Field.__slots__ + ('deep_copy',)
    type_repr: str = 'Any'
    atomic = False

    def __init__(self, *, deep_copy: bool = False, default: Any = Missing, hide_none: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing):
        super().__init__(default=default, hide_none=hide_none,
                         primitive_name=primitive_name, to_primitive_name=to_primitive_name)
        self.deep_copy = deep_copy

    def check_default(self):
        if not self.required and self.default_factory is None and self.default is None and not self.allow_none:
            raise ConfigurationError(f'Field `{self.name}` is not Optional and cannot use None as default')

    def convert(self, value: Any) -> Any:
        if value is Missing:
            return self._fill_missing()
        return deepcopy(value) if self.deep_copy else value

    def copy_value(self, value: Any) -> Any:
        return deepcopy(value)

    def to_primitive(self, value: Any) -> Any:
        return deepcopy(value) if self.deep_copy else value



from __future__ import annotations

from typing import Dict, Any, Iterable, Tuple, get_type_hints, cast, List, Union, Set, Type, TYPE_CHECKING

from stereotype.fields.base import Field, AnyField
from stereotype.fields.compound import ListField, DictField
from stereotype.fields.serializable import SerializableField
from stereotype.roles import Role, RequestedRoleFields, FinalizedRoleFields
from stereotype.utils import ConfigurationError

if TYPE_CHECKING:  # pragma: no cover
    from stereotype.model import Model


class ModelMeta(type):
    def __new__(mcs, name: str, bases: Tuple[type, ...], attrs: Dict[str, Any]):
        if not bases:
            return type.__new__(mcs, name, bases, attrs)

        # Only annotated attributes iterated here for the purposes of slots, not serializable fields
        field_names = [name for name, annotation in mcs._iterate_fields(attrs.get('__annotations__', {}))]
        field_values = {}
        for field_name in field_names:
            if field_name in attrs:
                # Field attributes popped as they will become instance attributes instead of class attributes
                field_values[field_name] = attrs.pop(field_name)

        # Find serializable, keep them in `attrs` as they need to remain as properties
        serializable_names = {name for name, _ in mcs._iterate_serializable(attrs)}

        # Using dicts instead of sets to preserve order
        all_slots = {
            **{slot: 0 for parent in bases for slot in parent.__slots__ or getattr(parent, '__abstract_slots__', ())},
            **{slot: 1 for slot in field_names},
            **{slot: 2 for slot in attrs.get('__slots__', ())},
        }
        if attrs.pop('__abstract__', False):
            attrs['__abstract_slots__'] = all_slots
        else:
            attrs['__slots__'] = [name for name in all_slots if name not in serializable_names]
        attrs['__fields__'] = NotImplemented
        attrs['__input_fields__'] = NotImplemented
        attrs['__validated_fields__'] = NotImplemented
        attrs['__role_fields__'] = NotImplemented
        attrs['__roles__'] = NotImplemented

        own_field_names = set(field_names) | serializable_names

        try:
            cls = cast(Type['Model'], type.__new__(mcs, name, bases, attrs))
        except TypeError as e:
            raise ConfigurationError(f'{name}: {e}, if inheriting from multiple models, only one may have __slots__ '
                                     f'(declare abstract models without __slots__ by adding class attribute '
                                     f'`__abstract__ = True`)')
        cls.__initialize_model__ = lambda *_: mcs._initialize_model(cls, bases, own_field_names, field_values)
        return cls

    @classmethod
    def _iterate_fields(mcs, field_annotations: Dict[str, Any]) -> Iterable[Tuple[str, Any]]:
        for name, annotation in field_annotations.items():
            if name.startswith('_'):
                continue
            yield name, annotation

    @classmethod
    def _initialize_model(mcs, cls: Type[Model], bases: Tuple[type, ...],
                          own_field_names: Set[str], field_values: dict):
        from stereotype.model import Model
        model_bases = [base for base in bases if issubclass(base, Model) and base is not Model]
        mcs._ensure_parent_models(model_bases, own_field_names, field_values)

        serializable = list(mcs._analyze_serializable(cls))
        serializable_names = {field.name for field in serializable}
        input_fields = [field for field in mcs._analyze_fields(cls, field_values)
                        if field.name not in serializable_names]
        cls.__input_fields__ = [field.make_input_config() for field in input_fields]
        cls.__validated_fields__ = [field.make_validated_config() for field in input_fields if field.has_validation()]
        cls.__fields__ = input_fields + serializable
        mcs._build_roles(cls, model_bases, own_field_names)

    @classmethod
    def _ensure_parent_models(mcs, model_bases: List[Type[Model]], own_field_names: Set[str], field_values: dict):
        for base in model_bases:
            if base.__fields__ is NotImplemented:
                base.__initialize_model__()
            for field in cast(List[Field], base.__fields__):
                if field.name not in own_field_names:
                    field_values[field.name] = field

    @classmethod
    def _analyze_fields(mcs, cls: Type[Model], field_values: dict) -> Iterable[Field]:
        for name, annotation in mcs._iterate_fields(mcs._resolve_annotations(cls)):
            analyzed_field = mcs._analyze_annotation(annotation)
            value = field_values.get(name)
            if isinstance(value, Field):
                field = value.copy_field()
                if not isinstance(field, type(analyzed_field)):
                    field.fill_in_name(name)
                    analyzed_field.fill_in_name(name)
                    raise ConfigurationError(f'Annotations for {analyzed_field} require custom field type '
                                             f'{type(analyzed_field).__name__}, got {type(field).__name__}')
                field.type_config_from(analyzed_field)
            else:
                field = analyzed_field
                if name in field_values:
                    field.set_default(value)

            field.fill_in_name(name)
            field.check_default()

            validator_method = getattr(cls, f'validate_{name}', None)
            if validator_method is not None:
                field.validator_method = validator_method

            yield field

    @classmethod
    def _iterate_serializable(mcs, attrs: Dict[str, Any]) -> Iterable[Tuple[str, SerializableField]]:
        for name, attr in attrs.items():
            if not isinstance(attr, property) or not hasattr(attr.fget, '__field__'):
                continue
            field = getattr(attr.fget, '__field__')
            if isinstance(field, SerializableField):
                yield name, field

    @classmethod
    def _analyze_serializable(mcs, cls: Type[Model]):
        attrs = {name: getattr(cls, name) for name in dir(cls)}
        for name, field in mcs._iterate_serializable(attrs):
            field.fill_in_name(name)
            yield field

    @staticmethod
    def _resolve_annotations(cls: Type[Model]) -> Dict[str, Any]:
        extra_types: Set[Type[Model]] = cls.resolve_extra_types()
        extra_locals = {typ.__name__: typ for typ in extra_types}
        try:
            return get_type_hints(cls, localns=extra_locals)
        except NameError as e:
            raise ConfigurationError(f'Model {cls.__name__} annotation {str(e)}. If not a global symbol or cannot be '
                                     f'imported globally, use the class method `resolve_extra_types` to provide it.')

    @classmethod
    def _analyze_annotation(mcs, annotation) -> Field:
        """Any supported annotation -> any Field"""
        from stereotype.model import Model

        typing_repr = repr(annotation)
        if typing_repr.startswith('typing.'):
            if typing_repr.startswith('typing.List['):
                return mcs._analyze_annotation_list(annotation)
            if typing_repr.startswith('typing.Dict['):
                return mcs._analyze_annotation_dict(annotation)
            if typing_repr.startswith('typing.Union['):
                return mcs._analyze_annotation_union(annotation)
            if typing_repr == 'typing.Any':
                return AnyField()
        else:
            if annotation in (bool, int, float, str):
                return mcs._analyze_annotation_atomic(annotation)
            if issubclass(annotation, Model):
                return mcs._analyze_annotation_model(annotation)

        raise ConfigurationError(f'Unsupported Model field {annotation!r}')

    @classmethod
    def _analyze_annotation_atomic(mcs, annotation) -> Field:
        """Atomic types -> atomic Field"""
        from stereotype.fields.atomic import ATOMIC_TYPE_MAPPING
        return ATOMIC_TYPE_MAPPING[annotation]()

    @classmethod
    def _analyze_annotation_model(mcs, annotation) -> Field:
        """Model subclass -> ModelField"""
        from stereotype import ModelField
        field = ModelField()
        field.type = annotation
        return field

    @classmethod
    def _analyze_annotation_union(mcs, annotation) -> Field:
        """Optional or Union -> any Field with allow_none or DynamicModelField"""
        from stereotype.model import Model

        options = annotation.__args__
        non_none = [option for option in options if option not in (type(None),)]

        if len(non_none) < len(options):
            # Optional is always represented by Union[None, ...]
            return mcs._analyze_annotation_optional(non_none)
        elif all(issubclass(option, Model) for option in options):
            return mcs._analyze_annotation_model_union(annotation, options)
        else:
            raise ConfigurationError(f'Union Model fields can only be Optional or Union of Model subclass types, '
                                     f'got {annotation!r}')

    @classmethod
    def _analyze_annotation_optional(mcs, non_none: List[Any]) -> Field:
        """Optional[...] -> any Field with allow_none"""
        optional_type = non_none[0] if len(non_none) == 1 else Union[tuple(non_none)]
        field = mcs._analyze_annotation(optional_type)
        field.allow_none = True
        return field

    @classmethod
    def _analyze_annotation_model_union(mcs, annotation, options: List[Any]) -> Field:
        """Union[ModelSubClass, ...] -> DynamicModelField"""
        from stereotype import DynamicModelField

        type_map = {}
        for option in options:
            if not hasattr(option, 'type'):
                raise ConfigurationError(f"Model {option.__name__} used in a dynamic model field {annotation} but does "
                                         f"not define a non-type-annotated string `type` field")
            if type(option.type).__name__ == 'member_descriptor':
                raise ConfigurationError(f"Model {option.__name__} used in a dynamic model field {annotation} but it's "
                                         f"`type` field has a type annotation making it a field, must be an attribute")
            if not isinstance(option.type, str):
                raise ConfigurationError(f"Model {option.__name__} used in a dynamic model field {annotation} but it's "
                                         f"`type` field {option.type} is not a string")
            if option.type in type_map:
                raise ConfigurationError(f"Conflicting dynamic model field types in {annotation}: "
                                         f"{type_map[option.type].__name__} vs {option.__name__}")
            type_map[option.type] = option

        field = DynamicModelField()
        field.type_map = type_map
        field.types = tuple(type_map.values())
        return field

    @classmethod
    def _analyze_annotation_list(mcs, annotation) -> Field:
        """List[...] -> ListField"""
        field = ListField()
        item_annotation, = annotation.__args__
        field.item_field = mcs._analyze_annotation(item_annotation)
        return field

    @classmethod
    def _analyze_annotation_dict(mcs, annotation) -> Field:
        """Dict[..., ....] -> DictField"""
        field = DictField()
        key_annotation, value_annotation = annotation.__args__
        field.key_field = mcs._analyze_annotation(key_annotation)
        if not field.key_field.atomic:
            raise ConfigurationError(f'DictField keys may only be booleans, numbers or strings: {annotation}')
        field.value_field = mcs._analyze_annotation(value_annotation)
        return field

    @classmethod
    def _build_roles(mcs, cls: Type[Model], bases: List[Type[Model]], own_field_names: Set[str]):
        roles = mcs._collect_finalized_roles(cls, bases, own_field_names)
        max_role_code = max((role.code for role in roles.keys()), default=0)
        default_role_fields = [field.make_output_config() for field in cls.__fields__]
        cls.__role_fields__ = [default_role_fields] * (max_role_code + 1)
        cls.__roles__ = []

        for role, finalized in roles.items():
            field_configs = [field.make_output_config() for field in cls.__fields__ if field.name in finalized.fields]
            cls.__role_fields__[role.code] = field_configs
            cls.__roles__.append(finalized)

    @classmethod
    def _collect_finalized_roles(mcs, cls: Type[Model], bases: List[Type[Model]], own_field_names: Set[str],
                                 ) -> Dict[Role, FinalizedRoleFields]:
        all_field_names, own_requested_roles = mcs._collect_own_requested_roles(cls)
        all_roles = {finalized.role for base in bases for finalized in base.__roles__} | set(own_requested_roles.keys())
        roles: Dict[Role, FinalizedRoleFields] = {role: FinalizedRoleFields(role) for role in all_roles}

        for base in reversed(bases):
            base_roles: Dict[Role, FinalizedRoleFields] = {finalized.role: finalized for finalized in base.__roles__}
            for role, finalized in roles.items():
                base_finalized = base_roles.get(role)
                if base_finalized is not None:
                    finalized.fields.update(base_finalized.fields)
                elif not role.empty_by_default:
                    finalized.fields.update(field.name for field in base.__fields__)

        for role, finalized in roles.items():
            own_requested = own_requested_roles.get(role)
            if own_requested is not None:
                finalized.update_requested(own_requested, all_field_names, own_field_names)
            elif not role.empty_by_default:
                finalized.fields.update(own_field_names)

        return roles

    @classmethod
    def _collect_own_requested_roles(mcs, cls: Type[Model]) -> Tuple[Set[str], Dict[Role, RequestedRoleFields]]:
        all_field_names = {field.name for field in cls.__fields__}
        own_requested_roles: Dict[Role, RequestedRoleFields] = {}
        for requested in cls.declare_roles():
            if requested.role in own_requested_roles:
                raise ConfigurationError(f'Role {requested.role.name} configured for {cls.__name__} multiple times')
            own_requested_roles[requested.role] = requested
        return all_field_names, own_requested_roles



from __future__ import annotations

from typing import Callable, Any, Optional, TYPE_CHECKING

from stereotype.fields.base import Field
from stereotype.utils import Missing

if TYPE_CHECKING:  # pragma: no cover
    from stereotype.model import Model


class SerializableField(Field):
    type_repr: str = 'serializable'

    def __init__(self, func: Callable[[Model], Any] = Missing, *,
                 hide_none: bool = False, to_primitive_name: str = Missing):
        super().__init__(hide_none=hide_none, to_primitive_name=to_primitive_name)
        self.serializable = func


def serializable(func: Optional[Callable[[Model], Any]] = None, *, hide_none: bool = False,
                 to_primitive_name: str = Missing):
    def serializable_wrapper(wrapped_func: Callable[[Model], Any]):
        field = SerializableField(wrapped_func, hide_none=hide_none, to_primitive_name=to_primitive_name)
        wrapped_func.__field__ = field
        return property(wrapped_func)

    if func is not None:
        assert callable(func), 'The positional argument `func` of `serializable` is only for usage without a call'
        return serializable_wrapper(func)

    return serializable_wrapper



from typing import Optional

from stereotype import Model, FloatField, ValidationError, Missing, ConversionError


# Keep this file synchronized with the `README.md` documentation!


class Employee(Model):
    name: str
    department: str = 'Engineering'
    female: Optional[bool] = None
    salary: float = FloatField(min_value=42., default=42.)

    human = True

    def greeting(self) -> str:
        title = {True: "Ms. ", False: "Mr. "}.get(self.female, "")
        return f'Dear {title}{self.name}'


model = Employee({'name': 'Jane Doe', 'female': True, 'salary': 150000})
model.validate()
assert model.name == 'Jane Doe'
assert model.department == 'Engineering'
assert model.female is True
assert model.salary == 150000.0
assert model.human is True
assert model.greeting() == 'Dear Ms. Jane Doe'
model.salary = 200000.0
primitive = model.to_primitive()
assert primitive == {'name': 'Jane Doe', 'department': 'Engineering', 'female': True, 'salary': 200000.0}

model = Employee({'department': None, 'salary': '10'})
try:
    model.validate()
    assert False, 'Validation fails due to errors below'  # pragma: no cover
except ValidationError as e:
    assert str(e) == 'name: This field is required'
    assert e.errors == {
        'name': ['This field is required'],
        'department': ['This field is required'],
        'salary': ['Must be at least 42.0'],
    }
assert model.name is Missing
assert model.department is None
assert model.salary == 10.0
assert model.to_primitive() == {'department': None, 'female': None, 'salary': 10.0}

try:
    model = Employee({'name': 'Alien', 'female': 'Unknown'})
except ConversionError as e:
    assert str(e) == 'female: Value must be a boolean or a true/false/yes/no string value'



from __future__ import annotations

import math
from typing import Optional, Iterable
from unittest import TestCase

from stereotype import Model, serializable, Role, RequestedRoleFields, IntField

role_a = Role('role_a')
role_b = Role('role_b')


class MyModel(Model):
    a: int
    b: int

    @serializable
    def a_plus_b(self) -> int:
        return self.a + self.b

    @serializable(hide_none=True)
    def lower_key(self) -> Optional[str]:
        if self.a == self.b:
            return None
        return 'A' if self.a < self.b else 'B'

    @serializable(to_primitive_name='c')
    def pythagoras(self) -> float:
        return math.sqrt(self.a * self.a + self.b * self.b)

    @property
    def a_minus_b(self) -> int:
        return self.a - self.b

    @classmethod
    def declare_roles(cls) -> Iterable[RequestedRoleFields]:
        yield role_a.blacklist(cls.b, cls.pythagoras)
        yield role_b.whitelist(cls.lower_key)


class TestSerializable(TestCase):
    def test_serializable_normal(self):
        model = MyModel({'a': 3, 'b': 4, 'a_plus_b': 'not 7'})
        self.assertEqual(3, model.a)
        self.assertEqual(4, model.b)
        self.assertEqual(7, model.a_plus_b)
        self.assertEqual('A', model.lower_key)
        self.assertEqual(5.0, model.pythagoras)
        self.assertEqual(-1, model.a_minus_b)
        self.assertEqual({'a': 3, 'b': 4, 'a_plus_b': 7, 'lower_key': 'A', 'c': 5.0}, model.serialize())
        self.assertEqual({'lower_key': 'A'}, model.serialize(role_b))
        model.validate()
        self.assertEqual('<Field pythagoras of type serializable, required, to output c>', str(MyModel.__fields__[-1]))
        self.assertEqual('<MyModel {a=3, b=4}>', repr(model))

    def test_serializable_special(self):
        model = MyModel({'a': 2, 'b': 2})
        self.assertEqual(2, model.a)
        self.assertEqual(2, model.b)
        self.assertEqual(4, model.a_plus_b)
        self.assertIsNone(model.lower_key)
        self.assertAlmostEqual(2.828, model.pythagoras, 3)
        self.assertEqual({'a': 2, 'a_plus_b': 4}, model.serialize(role_a))
        self.assertEqual({}, model.serialize(role_b))
        model.validate()

    def test_serializable_inheritance(self):
        class ChildModel(MyModel):
            @serializable
            def a_plus_b(self) -> int:
                return self.a * self.b

            @serializable
            def real_a_plus_b(self) -> int:
                return self.a + self.b

            @property
            def fake_field(self):
                return 42

            @classmethod
            def declare_roles(cls) -> Iterable[RequestedRoleFields]:
                yield role_b.blacklist(cls.real_a_plus_b)

        ChildModel.fake_field.fget.__field__ = IntField(default=47)

        model = ChildModel({'a': 12, 'b': -5})
        self.assertEqual(12, model.a)
        self.assertEqual(-5, model.b)
        self.assertEqual(-60, model.a_plus_b)
        self.assertEqual(7, model.real_a_plus_b)
        self.assertEqual('B', model.lower_key)
        self.assertEqual(42, model.fake_field)
        self.assertAlmostEqual(13.0, model.pythagoras, 3)
        self.assertEqual({'a': 12, 'b': -5, 'a_plus_b': -60, 'lower_key': 'B', 'c': 13.0, 'real_a_plus_b': 7},
                         model.serialize())
        self.assertEqual({'a': 12, 'a_plus_b': -60, 'lower_key': 'B', 'real_a_plus_b': 7}, model.serialize(role_a))
        self.assertEqual({'lower_key': 'B', 'a_plus_b': -60}, model.serialize(role_b))
        model.validate()

    def test_serializable_override(self):
        class First(Model):
            overridden_attribute: int
            inherited_attribute: int

            @serializable
            def overridden_serializable(self) -> int:
                return -self.overridden_attribute

            @serializable
            def inherited_serializable(self) -> int:
                return -self.inherited_attribute

        class Second(First):
            overridden_serializable: int

            @serializable
            def overridden_attribute(self) -> int:
                return -self.overridden_serializable

        model = Second({'overridden_attribute': 1, 'inherited_attribute': 2,
                        'overridden_serializable': 3, 'inherited_serializable': 1})
        self.assertEqual(-3, model.overridden_attribute)
        self.assertEqual(2, model.inherited_attribute)
        self.assertEqual(3, model.overridden_serializable)
        self.assertEqual(-2, model.inherited_serializable)
        self.assertEqual({'overridden_attribute': -3, 'inherited_attribute': 2,
                          'overridden_serializable': 3, 'inherited_serializable': -2}, model.serialize())
        model = First({'overridden_attribute': 1, 'inherited_attribute': 2})
        self.assertEqual(-1, model.overridden_serializable)
        self.assertEqual(-2, model.inherited_serializable)
        self.assertEqual({'overridden_attribute': 1, 'inherited_attribute': 2,
                          'overridden_serializable': -1, 'inherited_serializable': -2}, model.serialize())

    def test_items(self):
        model = MyModel({})
        self.assertEqual([], list(model.items()))
        model = MyModel({'a': 12, 'b': -5})
        self.assertEqual([('a', 12), ('b', -5)], list(model.items()))



from __future__ import annotations

from typing import Set, cast, Any, Optional, List, Dict, Union, Type
from unittest import TestCase

from stereotype import Model, Missing, ValidationError, ConversionError, BoolField, IntField, ConfigurationError, \
    FloatField, StrField, DataError
from stereotype.fields.base import Field, AnyField
from stereotype.fields.compound import DictField
from tests.common import Leaf


class DriedLeaf(Leaf):
    age: float = FloatField(max_value=5, default=0.)


class TestModels(TestCase):
    def test_bad_init(self):
        with self.assertRaises(ConversionError) as ctx:
            DriedLeaf(cast(dict, 'bad'))
        self.assertEqual('Supplied type str, needs a mapping', str(ctx.exception))
        self.assertEqual({'_global': ['Supplied type str, needs a mapping']}, ctx.exception.errors)

    def test_initialized_baseclass(self):
        Leaf()  # Ensures it is pre-initialized
        model = DriedLeaf({'color': 'brown'})
        self.assertEqual('brown', model.color)
        self.assertEqual(0, model.age)
        model.validate()
        self.assertEqual({'color': 'brown', 'age': 0}, model.serialize())

    def test_non_initialized_baseclass(self):
        class MyBase(Model):
            a: int
            b: float
            c: int = IntField(hide_none=True, default=-5)

        class MyChild(MyBase):
            b: str = StrField(primitive_name='bb')
            d: bool = True

            def validate_a(self, value, _):
                if value <= 0:
                    raise ValueError('Must be positive')

            def validate_b(self, value, _):
                if not value:
                    raise ValueError('Must be truthy')

            def validate_c(self, value, _):
                if value >= 0:
                    raise ValueError('Must be negative')

        model = MyChild({'a': 1, 'b': 'ignored', 'bb': 'abc'})
        self.assertEqual(1, model.a)
        self.assertEqual('abc', model.b)
        self.assertEqual(-5, model.c)
        self.assertIs(True, model.d)
        model.validate()
        self.assertEqual({'a': 1, 'bb': 'abc', 'c': -5, 'd': True}, model.serialize())
        model.a = 0
        model.b = ''
        model.c = 4
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual({'a': ['Must be positive'], 'bb': ['Must be truthy'], 'c': ['Must be negative']},
                         ctx.exception.errors)

        # If fields in bases were not copied, b would inherit the validator
        base = MyBase({'a': 0, 'b': 0, 'c': 1})
        self.assertEqual(0, base.a)
        self.assertEqual(0, base.b)
        self.assertEqual(1, base.c)
        base.validate()
        self.assertEqual({'a': 0, 'b': 0, 'c': 1}, base.serialize())

    def test_bad_field_type_typing(self):
        class BadType(Model):
            set: Set[int]

        with self.assertRaises(ConfigurationError) as ctx:
            BadType()
        self.assertEqual('Unsupported Model field typing.Set[int]', str(ctx.exception))

    def test_bad_field_type_native(self):
        class BadType(Model):
            bad: complex

        with self.assertRaises(ConfigurationError) as ctx:
            BadType()
        self.assertEqual("Unsupported Model field <class 'complex'>", str(ctx.exception))

    def test_multiple_non_abstract_bases(self):
        class Base1(Model):
            a: int

        class Base2(Model):
            b: int

        with self.assertRaises(ConfigurationError) as ctx:
            class Parent(Base1, Base2):
                unused: Parent
        self.assertEqual('Parent: multiple bases have instance lay-out conflict, if inheriting from multiple models, '
                         'only one may have __slots__ (declare abstract models without __slots__ by adding class '
                         'attribute `__abstract__ = True`)', str(ctx.exception))

    def test_none_primitive_name(self):
        class NonePrimitive(Model):
            normal: bool = BoolField(primitive_name='ok', to_primitive_name='ordinary')
            none_both: int = IntField(primitive_name=None)
            none_input: float = FloatField(primitive_name=None, to_primitive_name='no_input')
            none_output: str = StrField(to_primitive_name=None)

        model = NonePrimitive({'ok': True, 'none_both': 1, 'none_input': 4.2, 'no_input': 4.7, 'none_output': 'data'})
        self.assertIs(True, model.normal)
        self.assertIs(Missing, model.none_both)
        self.assertIs(Missing, model.none_input)
        self.assertEqual('data', model.none_output)
        self.assertEqual({'ordinary': True}, model.serialize())

        model.none_both = 5
        model.none_input = 42.47
        model.none_output = 'changed'
        self.assertEqual({'ordinary': True, 'no_input': 42.47}, model.serialize())
        self.assertEqual(5, model.none_both)
        self.assertEqual('changed', model.none_output)

    def test_equality(self):
        model = DriedLeaf({'age': 4.2})
        self.assertEqual(DriedLeaf({'color': 'green', 'age': 4.2}), model)
        self.assertNotEqual(DriedLeaf({'color': 'green', 'age': 4.7}), model)
        self.assertNotEqual(DriedLeaf({'color': None, 'age': 4.2}), model)
        self.assertNotEqual(Leaf({'color': 'green', 'age': 4.2}), model)
        self.assertNotEqual(1, model)
        self.assertNotEqual(None, model)

    def test_override_field_type(self):
        class First(Model):
            plain: int
            custom: int = IntField(min_value=1)

        class Second(First):
            plain: float
            custom: float = FloatField(min_value=2)

        self.assertEqual({'plain': 1.1, 'custom': 2.2}, Second({'plain': '1.1', 'custom': '2.2'}).serialize())

    def test_resolve_extra_types_error(self):
        class Inner(Model):
            pass

        class Outer(Model):
            field: Inner

        with self.assertRaises(ConfigurationError) as e:
            Outer()
        self.assertEqual("Model Outer annotation name 'Inner' is not defined. If not a global symbol or cannot be "
                         "imported globally, use the class method `resolve_extra_types` to provide it.",
                         str(e.exception))

    def test_copy(self):
        class Dummy(Model):
            type = 'dummy'
            field: List[int] = list

        class CopyRight(Model):
            atomic: float
            any: Any
            atomic_list: List[str]
            complex_list: Optional[List[List[bool]]] = None
            atomic_dict: Dict[int, bool] = DictField(hide_none=True)
            complex_dict: Dict[bool, Dummy]
            model: Dummy
            dynamic_model: Optional[Union[Dummy, Leaf]]

            @classmethod
            def resolve_extra_types(cls) -> Set[Type[Model]]:
                return {Dummy}

        model = CopyRight({'any': None, 'atomic_dict': None, 'dynamic_model': None})
        copy = model.copy(deep=True)
        model.atomic = 4.2
        model.any = {1: {True: False}}
        model.atomic_list = ['x']
        model.complex_list = [[]]
        model.atomic_dict = {1: True}
        model.complex_dict = {True: Dummy({'field': [1]})}
        model.model = None
        model.dynamic_model = Dummy()
        self.assertEqual({'any': None, 'complex_list': None, 'dynamic_model': None}, copy.serialize())

        model.model = Dummy({'field': [1, 2]})
        deep_copy = model.copy(deep=True)
        shallow_copy = model.copy(deep=False)
        model.atomic = 4.7
        model.any[1][False] = True
        model.any = Missing
        model.atomic_list.pop(-1)
        model.complex_list[0].append(True)
        model.complex_list.extend([[False], [True, False]])
        model.atomic_dict.pop(1)
        model.complex_dict[True].field.append(2)
        model.model.field.clear()
        model.dynamic_model.field.append(3)
        self.assertEqual({
            'atomic': 4.7,
            'atomic_list': [],
            'complex_list': [[True], [False], [True, False]],
            'atomic_dict': {},
            'complex_dict': {True: {'field': [1, 2]}},
            'model': {'field': []},
            'dynamic_model': {'type': 'dummy', 'field': [3]},
        }, model.to_primitive())
        self.assertEqual({
            'atomic': 4.2,
            'any': {1: {True: False}},
            'atomic_list': ['x'],
            'complex_list': [[]],
            'atomic_dict': {1: True},
            'complex_dict': {True: {'field': [1]}},
            'model': {'field': [1, 2]},
            'dynamic_model': {'type': 'dummy', 'field': []},
        }, deep_copy.serialize())
        self.assertEqual({
            'atomic': 4.2,
            'any': {1: {True: False, False: True}},
            'atomic_list': [],
            'complex_list': [[True], [False], [True, False]],
            'atomic_dict': {},
            'complex_dict': {True: {'field': [1, 2]}},
            'model': {'field': []},
            'dynamic_model': {'type': 'dummy', 'field': [3]},
        }, shallow_copy.serialize())

    def test_any_field(self):
        class WithAny(Model):
            normal: Any
            optional: Optional[Any] = None
            custom: Optional[Any] = AnyField(deep_copy=True, default='x', hide_none=True, primitive_name='customized')

            def validate_optional(self, value, _):
                if value:
                    raise ValueError('Must be falsy')

        initial = {'normal': {'x': 1}, 'customized': [1, [], {}]}
        model = WithAny(initial)
        self.assertEqual({'x': 1}, model.normal)
        self.assertIsNone(model.optional)
        self.assertEqual([1, [], {}], model.custom)
        initial['customized'][1].append(2)
        initial['customized'].pop(-1)
        model.validate()
        self.assertEqual({'normal': {'x': 1}, 'optional': None, 'customized': [1, [], {}]}, model.serialize())

        model.normal['x'] = 'no copy'
        self.assertEqual({'normal': {'x': 'no copy'}, 'optional': None, 'customized': [1, [], {}]}, model.serialize())
        model.serialize()['customized'][1].append(3)
        self.assertEqual([1, [], {}], model.custom)

        model = WithAny({'optional': 123})
        self.assertEqual({'optional': 123, 'customized': 'x'}, model.serialize())
        with self.assertRaises(ValidationError) as e:
            model.validate()
        self.assertEqual({
            'normal': ['This field is required'],
            'optional': ['Must be falsy'],
        }, e.exception.errors)

    def test_any_field_configuration_error(self):
        class Bad(Model):
            bad: Any = None

        with self.assertRaises(ConfigurationError) as e:
            Bad()
        self.assertEqual('Field `bad` is not Optional and cannot use None as default', str(e.exception))

    def test_missing_coverage(self):
        # The only purpose of this test is to ensure 100% coverage for *dead* code, where possible
        self.assertEqual(1, IntField().to_primitive(1))

        class Temp(Model):
            field: int

        Temp()

        class FakeField(Field):
            atomic = False
            type = set

        fake_field = FakeField()
        fake_field.name = 'field'
        list(fake_field.validate(42, {}))
        fake_field.copy_value(42)
        Temp.__fields__[0] = fake_field

        with self.assertRaises(NotImplementedError):
            repr(Temp())

        with self.assertRaises(AssertionError):
            DataError([])
